from ppq.core import NetworkFramework, TargetPlatform
from ppq.parser import (
                        AscendExporter,
                        CaffeExporter,
                        CaffeParser,
                        EspdlExporter,
                        ExtensionExporter,
                        MNNExporter,
                        NativeExporter,
                        NativeImporter,
                        NCNNExporter,
                        NxpExporter,
                        OnnxExporter,
                        OnnxParser,
                        ONNXRUNTIMExporter,
                        OpenvinoExporter,
                        PPLBackendExporter,
                        PPLDSPCaffeExporter,
                        PPLDSPTICaffeExporter,
                        QNNDSPExporter,
                        SNPECaffeExporter,
                        TengineExporter,
                        TensorRTExporter_JSON,
)
from ppq.quantization.quantizer import (
                        AscendQuantizer,
                        EspdlInt16Quantizer,
                        EspdlQuantizer,
                        EspdlS3Quantizer,
                        EspdlS3Int16Quantizer,
                        ExtQuantizer,
                        FPGAQuantizer,
                        GraphCoreQuantizer,
                        MetaxChannelwiseQuantizer,
                        MetaxTensorwiseQuantizer,
                        MNNQuantizer,
                        NCNNQuantizer,
                        NXP_Quantizer,
                        OnnxruntimeQuantizer,
                        OpenvinoQuantizer,
                        PPL_DSP_Quantizer,
                        PPL_DSP_TI_Quantizer,
                        PPLCUDAQuantizer,
                        RKNN_PerTensorQuantizer,
                        TengineQuantizer,
                        TensorRTQuantizer,
                        TensorRTQuantizer_FP8,
)

__QUANTIZER_COLLECTION__ = {
    TargetPlatform.PPL_DSP_INT8: PPL_DSP_Quantizer,
    TargetPlatform.PPL_DSP_TI_INT8: PPL_DSP_TI_Quantizer,
    TargetPlatform.SNPE_INT8:    PPL_DSP_Quantizer,
    TargetPlatform.QNN_DSP_INT8: PPL_DSP_Quantizer,
    TargetPlatform.TRT_INT8:     TensorRTQuantizer,
    TargetPlatform.MNN_INT8:     MNNQuantizer,
    TargetPlatform.ASC_INT8:     AscendQuantizer,
    TargetPlatform.NCNN_INT8:    NCNNQuantizer,
    TargetPlatform.NXP_INT8:     NXP_Quantizer,
    TargetPlatform.RKNN_INT8:    RKNN_PerTensorQuantizer,
    TargetPlatform.METAX_INT8_C: MetaxChannelwiseQuantizer,
    TargetPlatform.METAX_INT8_T: MetaxTensorwiseQuantizer,
    TargetPlatform.PPL_CUDA_INT8: PPLCUDAQuantizer,
    TargetPlatform.EXTENSION:     ExtQuantizer,
    TargetPlatform.FPGA_INT8   :  FPGAQuantizer,
    TargetPlatform.OPENVINO_INT8: OpenvinoQuantizer,
    TargetPlatform.TENGINE_INT8:  TengineQuantizer,
    TargetPlatform.GRAPHCORE_FP8: GraphCoreQuantizer,
    TargetPlatform.TRT_FP8:       TensorRTQuantizer_FP8,
    TargetPlatform.ONNXRUNTIME:   OnnxruntimeQuantizer,
    TargetPlatform.ESPDL_INT8:      EspdlQuantizer,
    TargetPlatform.ESPDL_INT16:     EspdlInt16Quantizer,
    TargetPlatform.ESPDL_S3_INT8:  EspdlS3Quantizer,
    TargetPlatform.ESPDL_S3_INT16:  EspdlS3Int16Quantizer,
}


__PARSERS__ = {
    NetworkFramework.ONNX: OnnxParser,
    NetworkFramework.CAFFE: CaffeParser,
    NetworkFramework.NATIVE: NativeImporter
}


__EXPORTERS__ = {
    TargetPlatform.PPL_DSP_INT8:  PPLDSPCaffeExporter,
    TargetPlatform.PPL_DSP_TI_INT8: PPLDSPTICaffeExporter,
    TargetPlatform.QNN_DSP_INT8:  QNNDSPExporter,
    TargetPlatform.PPL_CUDA_INT8: PPLBackendExporter,
    TargetPlatform.SNPE_INT8:     SNPECaffeExporter,
    TargetPlatform.NXP_INT8:      NxpExporter,
    TargetPlatform.ONNX:          OnnxExporter,
    TargetPlatform.ONNXRUNTIME:   ONNXRUNTIMExporter,
    TargetPlatform.OPENVINO_INT8: OpenvinoExporter,
    TargetPlatform.CAFFE:         CaffeExporter,
    TargetPlatform.NATIVE:        NativeExporter,
    TargetPlatform.EXTENSION:     ExtensionExporter,
    TargetPlatform.RKNN_INT8:     OnnxExporter,
    TargetPlatform.METAX_INT8_C:  ONNXRUNTIMExporter,
    TargetPlatform.METAX_INT8_T:  ONNXRUNTIMExporter,
    TargetPlatform.TRT_INT8:      TensorRTExporter_JSON,
    TargetPlatform.ASC_INT8:      AscendExporter,
    TargetPlatform.TRT_FP8:       ONNXRUNTIMExporter,
    TargetPlatform.NCNN_INT8:     NCNNExporter,
    TargetPlatform.TENGINE_INT8:  TengineExporter,
    TargetPlatform.MNN_INT8:      MNNExporter,
    TargetPlatform.ESPDL_INT8:      EspdlExporter,
    TargetPlatform.ESPDL_INT16:     EspdlExporter,
    TargetPlatform.ESPDL_S3_INT8:  EspdlExporter,
    TargetPlatform.ESPDL_S3_INT16:  EspdlExporter,
}


__all__ = ['__QUANTIZER_COLLECTION__', '__PARSERS__', '__EXPORTERS__']



from tests.tmodel  import *
from tests.tscheme import *
from ..ppq     import *
from ..ppq.api import *

DEVICE = 'cuda'

with ENABLE_CUDA_KERNEL():
    for scheme in TEST_SCHEMES:
        for case in TORCH_TEST_BLOCKS:
            try:
                print(f'PPQ System test start with model {case.model_name}, Scheme: {scheme.name}')

                dataset = [case.input_generator().to(DEVICE) for _ in range(8)]
                model = case.model_builder().to(DEVICE).eval()
                reference_outputs = torch.cat([model(batch) for batch in dataset])

                quantized = quantize_torch_model(
                    model=model, 
                    calib_dataloader=dataset, 
                    calib_steps=8,
                    input_shape=case.input_generator().shape,
                    platform=scheme.quant_platform,
                    setting=scheme.setting,
                    verbose=False)

                executor = TorchExecutor(quantized)
                for op in quantized.operations.values():
                    if isinstance(op, QuantableOperation): op.dequantize()
                ppq_outputs = torch.cat([executor(batch)[0] for batch in dataset])

                for op in quantized.operations.values():
                    if isinstance(op, QuantableOperation): op.restore_quantize_state()
                quant_outputs = torch.cat([executor(batch)[0] for batch in dataset])
                assert torch_snr_error(ppq_outputs, reference_outputs).item() < 1e-4, (
                    f'Network Simulating Failed, expect error < 1e-3, '
                    f'got {torch_snr_error(ppq_outputs, reference_outputs).item()}')
                assert torch_snr_error(quant_outputs, reference_outputs).item() < 0.1, (
                    f'Network Quantization Failed, expect error < 0.1, '
                    f'got {torch_snr_error(quant_outputs, reference_outputs).item()}')
                
                if (case.depoly_platforms is None or 
                    scheme.export_platform in case.depoly_platforms):
                    export_ppq_graph(
                        graph=quantized, 
                        platform=scheme.export_platform, 
                        graph_save_to='tworkingspace/export',
                        config_save_to='tworkingspace/export.json')
            except NotImplementedError as e:
                pass


"""
                                                                                                     
                                                              ...::iirii:..                                             
                P7Ysi.  iS7u.                ..:::irri:rPBBBBBBBBBBBBBBBBBBBBBBBg57.                                    
                B   :Js:B7 .PS        :sEBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBDv                                
                B      .:     u  :5BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB2                             
                DQ            SBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBQBBBBBBBBBBBBBBBBBBBBBBBBBBRi                          
           JZ  ..i:.      .dBBBBBBBBBBBBBBBBBBBBBBBg7          .BQBBBBBBBBBBBBBBBBBBBBBBBBBBBBBD7                       
            YQ.         vBBBBBBBBBBBBBBBBBBBBBBMY.              QBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB7                    
              2D      YBBBBBBBBBBBBBBBBBBBBZ7                   BBBBBBBBBBBBBBr SBBBBBBBBBBBBBBBBBBBBM:                 
               :Bi  UBBBBBBBBBBBBBBBBBPv.                       BBBBBBBBBBBBj    .BBBBBBBBBBBBBBBBBBBBBBY               
                 ZQBBBBBBBBBBBBBBE                             DBBBBBBBBBBS        .PBBBBBBBBBBBBBBBBBBBBBu             
                ZBBBBBBBBBBBBBBBZ                             EBBBBBBBBb:             :PBBBBBBBBBBBBBBBBBBBB            
             .iBBBBBBBBBBBBBBBBB          ..:i:.            .BBBBBQKr                     7gBBBBBBBBBBBBBBBBB           
             gBBBBBBBBBBBBBBBBQ      .irri::...           IR2r.                   rir7j17.    :LBBBBQRQBBBBBBB          
             BBBBBBBBBBBBBBBB:     rur.                 .i.                              iKg7     vBBBBBBBBBBB:         
            BBBBBBBBBBBBBBP.      Xr    ..:r7rrr.                              .::LKIvrr.   5Br     :BBBBBBBBB1         
           bBBBBBBBBBBZ7.            :Yr.     :dBBP                         .JYi.      :PBE   E       BBBBBBBBr        
          jBBBBBBBBBB              :Q7    .rv:   vBb                        Qi            BB:         :BBBBBBB.        
          BBBBBBBBBBr              B.   BBBBBBBBQ                                5BBBBBQi  dB          IBBBBQB5        
         :BBBBBBBBBB                   JBBBBBBBBB:                              BBBBBBBBB:               rBBBBB        
          BBBBBBBBB1                    XBBBBBBU                                ZBBBBBBBB.                EBBBB7MQ     
          BBBBBBBBB7                                                              .LU1r.                  iBBBb  JBv   
          BBBBBBBBg                  5D7                         ..                        .              :BBB7   rB    
          qB: gBBBK           i  : .  .Y21JJL7i.                 ..B:             ULi::iirr:..            .BBu  :  B7  
        gd Ui  QBBE            .   :.                              r.              .....    :i. :.        PBR     .B.  
       7B .    iBBB                                                                             .        .B 7r    BL   
       BB .     UBBv                                                                                     BP .   rBr    
       MB      q  .                                                                                     XS    rDj:      
       .BB    .::. :                  iLv7ri:.                                      ...                ..  :YUr         
         BB7     .:U                   ...irvUXPP27:                      ..:71SqP2Is7i              Ui::rr:          
          .BBg:                                  .rLr:r7i..        .:irrrii:..                      Bg                
             7QBE7.                                     .:r77v .rri:.                             rB7                 
                :JdDE277S.                                   ..r                                rDK                   
                      . rBBi                                                                .is5r                     
                          :gBEr                                                          :7jL:                        
                             :SQMSr.                                                 .:7r:.                            
                                 :YKPbur::..                                 ...irrqBq.                                 
                                      :KgBQU1UUJr:          i7777vsv777v777ri:...    rdQJ.                              
                                    7qj:                          ..              .     7QBr                            
                                  rP:                                                      bB            

Sensetime PPL Quant Tool(PPQ)

Do not modify codes within this directory if not necessary.
"""

from .common import *
from .config import PPQ_CONFIG
from .data import (DataType, OperationMeta, TensorMeta, convert_any_to_numpy,
                   convert_any_to_python_primary_type, convert_any_to_string,
                   convert_any_to_torch_tensor, convert_primary_type_to_list)
from .defs import (SingletonMeta, empty_ppq_cache, ppq_debug_function,
                   ppq_file_io, ppq_info, ppq_legacy,
                   ppq_quant_param_computing_function, ppq_warning)
from .ffi import CUDA
from .quant import (ChannelwiseTensorQuantizationConfig, NetworkFramework,
                    OperationQuantizationConfig, QuantizationPolicy,
                    QuantizationProperty, QuantizationStates, RoundingPolicy,
                    TargetPlatform, TensorQuantizationConfig, QuantizationVisibility)
from .storage import (Serializable, ValueState, is_file_exist,
                      open_txt_file_from_writing)
from typing import Any, Callable, List, Iterable, Set, Dict, Union, Text

print(
r"""
    ___________ ____        ____  ____  ____ 
   / ____/ ___// __ \      / __ \/ __ \/ __ \
  / __/  \__ \/ /_/ /_____/ /_/ / /_/ / / / /
 / /___ ___/ / ____/_____/ ____/ ____/ /_/ / 
/_____//____/_/         /_/   /_/    \___\_\ 

""")



from typing import List

from ppq.core import (
    OperationQuantizationConfig,
)
from ppq.IR import BaseGraph, Operation, OperationExporter, Variable
from ppq.IR.quantize import QuantableOperation
from ppq.log import NaiveLogger
from ppq.parser.espdl.espdl_typedef import (
    ADD_LIKE_OP_SET,
    CONV_LAYOUT_OP_SET,
    OTHER_OP_SET,
    PASSIVE_LAYOUT_OP_SET,
    SOFTMAX_LIKE_OP_SET,
    ExporterPatternInfo,
)
from ppq.parser.espdl.export_patterns import fuse_downstream_operation

logger = NaiveLogger.get_logger('ESPDL')
# logger.set_level("DEBUG")


def transpose_shape(input_shape, perm: List[int]) -> List[int]:
    if not perm:
        return input_shape
    return [input_shape[i] for i in perm]


def get_inverse_transpose(perm: List[int]) -> List[int]:
    """
    tensor == inverse_transpose(transpose(tensor))
    """
    # return perm
    return [perm.index(i) for i in range(len(perm))]

def get_default_perm(var: Variable) -> List[int]:
    """
    return the default permute for given variable, [0,1,2,3,...]
    """
    if not var or not var.shape:
        return []

    return [i for i in range(len(var.shape))]


def insert_transpose_node(
    graph: BaseGraph, var: Variable, op: Operation, perm: List[int]
) -> Operation:
    """
    Insert a Transpose Node on given variable, according to given perm.
    """
    info = ExporterPatternInfo()

    if perm != range(len(perm)):
        logger.debug(
            f"insert transpose node: op: {op.name}, var:{var.name}, perm:{perm}"
        )
        created = graph.create_operation(op_type="Transpose", attributes={"perm": perm})
        if var in op.inputs:
            var_index = op.inputs.index(var)
            if isinstance(op, QuantableOperation):
                config = op.config
                # For transpose op,  input_quantization_config == output_quantization_config
                new_config = OperationQuantizationConfig(
                    [config.input_quantization_config[var_index]],
                    [config.input_quantization_config[var_index]],
                )
                created = QuantableOperation(created, new_config, op.platform)
                graph.operations[created.name] = created

            graph.insert_op_before(A=created, B=op, input_idx=var_index)
            new_var = created.outputs[0]
            new_var.shape = var.shape
            new_var.is_parameter = var.is_parameter
            new_var.dtype = var.dtype
            perm = get_default_perm(created.outputs[0])
            info.add_var_permute(
                created.outputs[0].name, get_default_perm(created.outputs[0])
            )

        else:
            raise ValueError(f"Unexpected Error in Exporting Op {op.name}({op.type}).")

        return created


def restore_origin_shape(op: Operation, graph: BaseGraph):
    info = ExporterPatternInfo()
    for var in op.inputs:
        if var.is_parameter:
            continue

        var_perm = info.get_var_permute(var.name)
        if var_perm and var_perm != get_default_perm(var):
            # There is already a permute, but this op need keep origin shape
            # A transpose node needs to be inserted into the word.
            inverse_perm = get_inverse_transpose(var_perm)
            insert_transpose_node(graph, var, op, inverse_perm)
        else:
            info.add_var_permute(var.name, get_default_perm(var))

    for var in op.outputs:
        info.add_var_permute(var.name, get_default_perm(var))
    return op


class ResetConvLayoutPattern(OperationExporter):
    """
    Modify Conv inputs and outputs layout from NCHW to NHWC
    And Update all variable's shape
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        info = ExporterPatternInfo()
        if op.type in CONV_LAYOUT_OP_SET:
            for var in op.inputs:
                if var.is_parameter:
                    continue

                var_shape = var.shape
                if len(var_shape) == 4:  # conv2d, NCHW -> NHWC
                    perm = [0, 2, 3, 1]
                elif len(var_shape) == 3:  # conv1d, NCW -> NWC
                    perm = [0, 2, 1]

                var_perm = info.get_var_permute(var.name)

                if var_perm:
                    if perm != var_perm:
                        # There is already a permute, but it does not match the conv layout.
                        # A transpose node needs to be inserted into the graph.
                        inverse_perm = get_inverse_transpose(var_perm)
                        new_perm = transpose_shape(inverse_perm, perm)
                        insert_transpose_node(graph, var, op, new_perm)
                else:
                    info.add_var_permute(var.name, perm)

            for var in op.outputs:
                var_shape = var.shape
                if len(var_shape) == 4:  # conv2d, NCHW -> NHWC
                    perm = [0, 2, 3, 1]
                else:  # conv1d, NCW -> NWC
                    perm = [0, 2, 1]
                info.add_var_permute(var.name, perm)

        return op


class RestoreOriginLayoutPattern(OperationExporter):
    """
    Restore original layout
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in OTHER_OP_SET:
            return restore_origin_shape(op, graph)

        return op


class BypassPassiveLayoutPattern(OperationExporter):
    """
    Passive Node inherit transpose from upstream
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in PASSIVE_LAYOUT_OP_SET:
            info = ExporterPatternInfo()
            assert len(op.outputs) == 1
            input1 = op.inputs[0]
            output = op.outputs[0]
            assert input1.shape == output.shape

            var_perm = info.get_var_permute(input1.name)
            if not var_perm:
                var_perm = get_default_perm(input1)
                info.add_var_permute(input1.name, var_perm)

            info.add_var_permute(op.outputs[0].name, var_perm)
        return op


class BypassAddLikePattern(OperationExporter):
    """
    Add,Mul,Sub,Div:

    two input and one output,
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ADD_LIKE_OP_SET:
            info = ExporterPatternInfo()
            input1 = op.inputs[0]
            input2 = op.inputs[1]
            output = op.outputs[0]
            input1_perm = info.get_var_permute(input1.name)
            input2_perm = info.get_var_permute(input2.name)
            output_perm = None
            if not input1.is_parameter and not input2.is_parameter:
                if input1_perm == input2_perm:
                    if input1_perm:
                        # using upstream's perm
                        output_perm = input1_perm
                    else:
                        # input1_perm is None, add new perm
                        info.add_var_permute(input1.name, get_default_perm(input1))
                        info.add_var_permute(input2.name, get_default_perm(input2))
                        output_perm = get_default_perm(output)
                    # logger.debug(f"{info.get_var_permute(input1.name)}, {info.get_var_permute(input2.name)}, {output_perm}")
                    info.add_var_permute(output.name, output_perm)
                else:
                    # insert transpose node and restore origin shape
                    return restore_origin_shape(op, graph)
            elif input2.is_parameter or input1.is_parameter:
                return restore_origin_shape(op, graph)

        return op

class BypassSoftmaxLayoutPattern(OperationExporter):
    """
    Softmax pattern with one input and one output and axis attribute
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in SOFTMAX_LIKE_OP_SET:
            info = ExporterPatternInfo()
            input = op.inputs[0]
            output = op.outputs[0]

            var_perm = info.get_var_permute(input.name)
            if var_perm and var_perm != get_default_perm(input):
                # There is already a permute, change axis accordingly
                axis = (int(op.attributes["axis"]) + len(var_perm)) % len(var_perm)
                new_axis = var_perm.index(axis)
                op.attributes["axis"] = new_axis
            else:
                var_perm = get_default_perm(input)
                info.add_var_permute(input.name, var_perm)
            
            # use input perm
            info.add_var_permute(output.name, var_perm)

        return op


class ResetConcatPattern(OperationExporter):
    """
    Concat pattern with two input and one output,
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ["Concat"]:
            perm_dict = {}
            info = ExporterPatternInfo()

            for var in op.inputs:
                var_perm = info.get_var_permute(var.name)
                perm_str = str(var_perm)
                if perm_str not in perm_dict:
                    perm_dict[perm_str] = var_perm

            output_var = op.outputs[0]

            if len(perm_dict) == 1:  # all input have same perm, output bypass
                var_perm = list(perm_dict.values())[0]
                if not var_perm:
                    restore_origin_shape(op, graph)
                else:
                    axis = op.attributes["axis"]
                    new_axis = var_perm.index(int(axis))
                    op.attributes["axis"] = new_axis
                    info.add_var_permute(output_var.name, var_perm)
                    logger.debug(f"{op.name} update axes from {axis} to {new_axis}")
            else:
                logger.debug(f"transpose perm {perm_dict}")
                restore_origin_shape(op, graph)
        return op


class ResetResizePattern(OperationExporter):
    """
    Reize Layout Pattern
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        if op.type in ["Resize"]:
            info = ExporterPatternInfo()
            input_var = op.inputs[0]
            if len(op.inputs) > 1:
                roi = op.inputs[1]
            else:
                roi = None
            
            if len(op.inputs) > 2:
                scales = op.inputs[2]
            else:
                scales = None
            
            if len(op.inputs) > 3:
                sizes = op.inputs[3]
            else:
                sizes = None
            
            for var in op.inputs[1:]:
                perm = info.get_var_permute(var.name, get_default_perm(var))

            input_perm = info.get_var_permute(input_var.name)
            output_var = op.outputs[0]

            if input_perm:
                for var in op.inputs[1:]:
                    if var:
                        info.add_var_permute(var.name, get_default_perm(var))
                info.add_var_permute(output_var.name, input_perm)
                # axes = op.attributes["axis"]
                # if axes:
                #     new_axes = [input_perm.index(i) for i in axes]
                #     op.attributes["axis"] = new_axes
                #     logger.debug(f"resize axes from {axes} to {new_axes} ")
                # else:
                #     if scales:
                #         values = scales.tolist()
                #         new_values = [
                #             values[input_perm[int(i)]] for i in range(len(values))
                #         ]
                #         scales.value = torch.Tensor(new_values, dtype=torch.int32)
                #         logger.debug(
                #             f"{op.name} reset scales from {values} to {new_values} "
                #         )
                #     if sizes:
                #         values = sizes.tolist()
                #         new_values = [
                #             values[input_perm[int(i)]] for i in range(len(values))
                #         ]
                #         sizes.value = torch.Tensor(new_values, dtype=torch.int32)
                #         logger.debug(
                #             f"{op.name} reset sizes from {values} to {new_values} "
                #         )
                #     if roi:
                #         values = sizes.tolist()
                #         new_values = []
                #         for i in range(len(values) / 2):
                #             new_values.append(values[input_perm[int(i)] * 2])
                #             new_values.append(values[input_perm[int(i)] * 2 + 1])
                #         roi.value = torch.Tensor(new_values, dtype=torch.int32)
                #         logger.debug(
                #             f"{op.name} reset sizes from {values} to {new_values} "
                #         )
            else:
                if len(input_var.shape) == 4:  # conv2d, NCHW -> NHWC
                    perm = [0, 2, 3, 1]
                elif len(input_var.shape) == 3:  # conv1d, NCW -> NWC
                    perm = [0, 2, 1]
                else:
                    logger.error(f"Reize: do not support shape for {input_var.shape}")
                
                info.add_var_permute(input_var.name, perm)
                info.add_var_permute(output_var.name, perm)

        return op





class FuseTransposePattern(OperationExporter):
    """
    Fuse Transpose Pattern
    """

    def export(self, op: Operation, graph: BaseGraph, **kwargs) -> Operation:
        # The FUSE_OP_PATTERNS may remove some ops.
        if op.name not in graph.operations:
            return op

        if op.type == "Transpose":
            downstream_transpose_op = None
            perm = op.attributes["perm"]
            while True:
                downstream_op = graph.get_downstream_operations(op)
                # the downstream op have only one op and this op is Transpose
                if len(downstream_op) == 1 and downstream_op[0].type == "Transpose":
                    downstream_transpose_op = downstream_op[0]
                    perm = transpose_shape(perm, downstream_op[0].attributes["perm"])

                    if isinstance(op, QuantableOperation):
                        op_config = op.config
                        downstream_config = downstream_transpose_op.config
                        new_config = OperationQuantizationConfig(
                            op_config.input_quantization_config,
                            downstream_config.output_quantization_config,
                        )
                        op.config = new_config
                    graph = fuse_downstream_operation(
                        graph, downstream_transpose_op, keep_coherence=True
                    )
                    op.attributes["perm"] = perm
                else:
                    break

            perm = op.attributes["perm"]
            if perm == [i for i in range(len(perm))]:
                # Removed redundant transpose
                graph.remove_operation(op, keep_coherence=True)
        return op

def print_vars(op: Operation):
    logger.info(f"Op: {op.name}, {op.type}, {op.attributes}")
    for var in op.inputs:
        print("inputs:", var.name, var.shape)
    for var in op.outputs:
        print("outputs:", var.name, var.shape)
    
def reset_graph_layout(graph: BaseGraph):
    """
    Reset layout from NCHW -> NHWC
    """

    layout_patterns = [
        [CONV_LAYOUT_OP_SET, ResetConvLayoutPattern], 
        [PASSIVE_LAYOUT_OP_SET, BypassPassiveLayoutPattern],
        [ADD_LIKE_OP_SET, BypassAddLikePattern],
        [SOFTMAX_LIKE_OP_SET, BypassSoftmaxLayoutPattern],
        [["Concat"], ResetConcatPattern],
        [["Resize"], ResetResizePattern],
        [OTHER_OP_SET, RestoreOriginLayoutPattern]
    ]
        
    for op in graph.topological_sort():
        flag = 1
        for pattern in layout_patterns:
            if op.type in pattern[0]:
                pattern[1]().export(op, graph)
                flag = 0
                break
        if flag:
            logger.error(f"Can not reset {op.type}:{op.name} layout")
    
    # fuse transpose op
    pattern = FuseTransposePattern()
    for op in graph.topological_sort():
        pattern.export(op, graph)



from typing import Union

import torch
from ppq.api.setting import QuantizationSetting
from ppq.core import (OperationQuantizationConfig, QuantizationPolicy,
                      QuantizationProperty, QuantizationStates, RoundingPolicy,
                      TargetPlatform)
from ppq.executor import BaseGraphExecutor
from ppq.IR import BaseGraph, Operation
from ppq.quantization.optim import (NCNNFormatGemmPass,
                                    QuantizationOptimizationPipeline)

from .base import BaseQuantizer


class NCNNQuantizer(BaseQuantizer):
    def __init__(
        self, graph: BaseGraph
    ) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = - 127
        self._quant_max = + 127

    def build_prequant_pipeline(
        self, setting: QuantizationSetting, executor: BaseGraphExecutor) -> QuantizationOptimizationPipeline:
        pipeline = super().build_prequant_pipeline(setting, executor)
        pipeline.append_optimization_to_pipeline(NCNNFormatGemmPass(), at_front=True)
        return pipeline

    def init_quantize_config(
        self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy, rounding=self.rounding_policy,
            op=operation, num_of_bits=self._num_of_bits, exponent_bits=0,
            quant_max=self._quant_max, quant_min=self._quant_min,
            observer_algorithm='percentile'
        )

        if operation.type in {'Conv', 'Gemm'}:
            assert operation.num_of_input > 0, 'Seems you got a Computing layer with no parameters.'

            if operation.type == 'Conv':
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'

                group        = operation.attributes.get('group', 1)
                dilations    = operation.attributes.get('dilations', [1, 1])
                strides      = operation.attributes.get('strides', [1, 1])
                kernel_shape = operation.attributes.get('kernel_shape')
                if group == 1 and all([i == 1 for i in dilations]) and all([j == 1 for j in strides])\
                    and all([k == 3 for k in kernel_shape]):
                    base_quant_config.input_quantization_config[1].num_of_bits = 6
                    base_quant_config.input_quantization_config[1].quant_max   = +31
                    base_quant_config.input_quantization_config[1].quant_min   = -31

            elif operation.type == 'Gemm':
                assert operation.attributes.get('transB', 0) and operation.attributes.get('alpha', 1.0) == 1.0 \
                    and operation.attributes.get('beta', 1.0) == 1.0
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL +
                    QuantizationProperty.LINEAR +
                    QuantizationProperty.PER_CHANNEL
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'minmax'

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

            base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32
        return base_quant_config

    @ property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.NCNN_INT8

    @ property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @ property
    def quant_operation_types(self) -> set:
        return {
            'Conv', 'Gemm'
        }

    @ property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL +
            QuantizationProperty.LINEAR +
            QuantizationProperty.PER_TENSOR
        )

    @ property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO

    @ property
    def activation_fusion_types(self) -> set:
        """
        ncnn 只需要输入定点 不需要考虑激活函数融合

        Returns:
            set: _description_
        """
        return {}



import torch

from ppq.core.quant import TargetPlatform

class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.gemm_1 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_2 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_3 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_4 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_5 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_6 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_7 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_8 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_9 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_10 = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_J = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_Q = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_K = torch.nn.Linear(in_features=10, out_features=10)
        self.gemm_A = torch.nn.Linear(in_features=10, out_features=10)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.gemm_1(x)
        x = torch.relu(x)
        
        x2 = torch.relu(self.gemm_2(x))
        x3 = torch.relu(self.gemm_3(x))
        x4 = torch.relu(self.gemm_4(x))
        x5 = torch.relu(self.gemm_5(x))
        x6 = torch.relu(self.gemm_6(x))
        
        x2 = self.gemm_7(x2)
        x3 = self.gemm_8(x3)
        x4 = self.gemm_9(x4)
        x5 = self.gemm_10(x5)
        x6 = self.gemm_J(x6)
        
        x7 = torch.relu(self.gemm_Q(x))
        x7 = self.gemm_K(x7)
        
        x8 = torch.max_pool1d(x7, kernel_size=2)
        return torch.cat([x2, x3, x4, x5, x6, x7, x8], dim=-1)

model = MyModel().cuda()
model.forward(torch.zeros(size=[10, 10]).cuda())

from ppq.api import load_torch_model
from ppq.api import quantize_native_model
from ppq.api import QuantizationSettingFactory

graph = load_torch_model(model=model, sample=torch.zeros(size=[10, 10]).cuda())
s = QuantizationSettingFactory.default_setting()
s.lsq_optimization = True
s.lsq_optimization_setting.block_size = 4

quantize_native_model(
    model=graph, calib_dataloader=[torch.zeros(size=[10, 10])], input_shape=[10, 10],
    calib_steps=8, collate_fn=lambda x: x.cuda(), platform=TargetPlatform.TRT_INT8,
    setting=s)


import json
import torch
from typing import List

from ppq.core import (DataType, QuantizationStates,
                      QuantizationVisibility, NetworkFramework, ppq_warning)
from ppq.IR import BaseGraph, GraphExporter
from ppq.IR.quantize import QuantableOperation

from .onnx_exporter import OnnxExporter
from .caffe_exporter import CaffeExporter
from .util import convert_value


class QNNDSPExporter(GraphExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        activation_info, param_info = {}, {}
        topo_order =  graph.topological_sort()
        for operation in topo_order:
            if not isinstance(operation, QuantableOperation): continue
            for config, var in operation.config_with_variable:
                if not QuantizationStates.can_export(config.state):
                    raise PermissionError(
                        'Can not export quant config cause not all quantization configurations '
                        'have been correctly initialized(or some of them has been deactivated). '
                        f'Operation {operation.name} has an invalid quantization state({config.state}) '
                        f'at variable {var.name}.')

                if config.visibility == QuantizationVisibility.INTERNAL: continue
                if config.state in {
                    QuantizationStates.FP32,
                    QuantizationStates.SOI
                }: continue

                if var.source_op is not None and var.source_op.type in {"Softmax", "Sigmoid"}:
                    if config.dominated_by == config: # changeable.
                        # fix output range greater than 1
                        config.scale = torch.clamp(config.scale, max=1.0 / (config.quant_max - config.quant_min))

                if config.state == QuantizationStates.PASSIVE and var.name in activation_info: continue
                info =  [{
                            'bitwidth': config.num_of_bits,
                            'max'     : convert_value(config.scale * (config.quant_max - config.offset), True, DataType.FP32),
                            'min'     : convert_value(config.scale * (config.quant_min - config.offset), True, DataType.FP32),
                            'offset'  : convert_value(config.offset, True, DataType.INT32),
                            'scale'   : convert_value(config.scale, True, DataType.FP32)
                        }]
                if var.is_parameter:
                    param_info[var.name] = info
                else:
                    activation_info[var.name] = info

        exports = {
            'activation_encodings': activation_info,
            'param_encodings': param_info
        }

        with open(file=config_path, mode='w') as file:
            json.dump(exports, file, indent=4)


    def export(self, file_path: str, graph: BaseGraph, config_path: str = None, input_shapes: List[List[int]] = [[1, 3, 224, 224]]):
        if config_path is not None:
            self.export_quantization_config(config_path, graph)
        if graph._built_from == NetworkFramework.CAFFE:
            exporter = CaffeExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None, input_shapes=input_shapes)
        elif graph._built_from == NetworkFramework.ONNX:
            exporter = OnnxExporter()
            exporter.export(file_path=file_path, graph=graph, config_path=None)



from abc import ABCMeta
from typing import Dict, List

from ppq.core import (QuantizationStates, TensorQuantizationConfig,
                      ppq_debug_function)
from ppq.executor import QuantOPRuntimeHook
from ppq.IR import QuantableOperation, Variable

from .base import BaseTensorObserver
from .order import TorchIsotoneObserver
from .range import (TorchHistObserver, TorchMinMaxObserver, TorchMSEObserver,
                    TorchPercentileObserver)
from .floating import ConstantObserver, DirectMSEObserver

OBSERVER_TABLE = {
    'minmax': TorchMinMaxObserver,
    'kl': TorchHistObserver,
    'percentile': TorchPercentileObserver,
    'mse': TorchMSEObserver,
    'isotone': TorchIsotoneObserver,
    'constant': ConstantObserver,
    'floating': DirectMSEObserver
}

class TensorObserverFactroy():
    def __init__(self) -> None:
        raise NotImplementedError(
            'Observer Factory can not be initialized, use TensorObserverFactroy.build_observer instead.')

    @ classmethod
    def build_observer(cls, variable: Variable, config: TensorQuantizationConfig) -> BaseTensorObserver:
        algorithm = str(config.observer_algorithm.lower())
        if algorithm not in OBSERVER_TABLE:
            raise ValueError(
                f'Observer type not understand, Except one of {OBSERVER_TABLE.keys()}, '\
                f'while {str(algorithm)} was given.')
        return OBSERVER_TABLE[algorithm](watch_on=variable, quant_cfg=config)


class CalibrationHook(QuantOPRuntimeHook):
    def __init__(
        self,
        operation: QuantableOperation,
        observer_table: Dict[Variable, BaseTensorObserver]
    ) -> None:
        self._operation = operation
        self._observer_table = observer_table
        super().__init__(operation)

    def pre_forward_hook(
        self, inputs: list, quant_inputs: list, quant_configs: List[TensorQuantizationConfig]) -> list:
        for input_var, quant_config in zip(inputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(input_var)
        return quant_inputs

    def post_forward_hook(
        self, outputs: list, quant_outputs: list, quant_configs: List[TensorQuantizationConfig]) -> list:
        for output_var, quant_config in zip(outputs, quant_configs):
            if quant_config in self._observer_table:
                observer = self._observer_table[quant_config]
                observer.observe(output_var)
        return quant_outputs

    def render_quantization_config(self):
        for _, observer in self._observer_table.items():
            observer.render_quantization_config()
            observer.report()

    def __str__(self) -> str:
        return ''.join([observer.__str__() + '\n' for _, observer in self._observer_table.items()])


class OperationObserver(metaclass=ABCMeta):
    def __init__(
        self,
        operation: QuantableOperation,
        monitor_parameter: bool = True,
        monitor_outputs  : bool = True,
        monitor_inputs   : bool = True
    ) -> None:

        if not isinstance(operation, QuantableOperation):
            raise TypeError(f'Only QuantableOP instance can apply an Observer, '\
                f'while {type(operation)} was given.')

        self._operation = operation
        self._hook = self.build_hook(
            monitor_parameter = monitor_parameter,
            monitor_outputs   = monitor_outputs,
            monitor_inputs    = monitor_inputs
        )

    def render_quantization_config(self):
        self.hook.render_quantization_config()

    def build_hook(self, monitor_parameter: bool,
        monitor_outputs: bool, monitor_inputs: bool) -> CalibrationHook:
        assert isinstance(self._operation, QuantableOperation)
        observer_table = {}
        for var, config in zip(
            self._operation.inputs, self._operation.config.input_quantization_config):
            if config.state == QuantizationStates.INITIAL:
                if var in self._operation.parameters and monitor_parameter:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)
                elif monitor_inputs:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)

        if monitor_outputs:
            for var, config in zip(
                self._operation.outputs, self._operation.config.output_quantization_config):
                if config.state == QuantizationStates.INITIAL:
                    observer_table[config] = TensorObserverFactroy.build_observer(var, config)

        return CalibrationHook(operation=self._operation, observer_table=observer_table)

    @ property
    def hook(self) -> CalibrationHook:
        return self._hook

    @ ppq_debug_function
    def report(self) -> str:
        return str(self._hook)



from .imagenet_util import (evaluate_mmlab_module_with_imagenet,
                            evaluate_onnx_module_with_imagenet,
                            evaluate_ppq_module_with_imagenet,
                            evaluate_torch_module_with_imagenet, 
                            load_imagenet_from_directory)



import collections.abc
import numbers
import os
import random
import struct
import sys
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import flatbuffers
import numpy as np
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

# Generated by `flatc`
sys.path.append(os.path.dirname(__file__))
import FlatBuffers.Dl.Attribute as Attribute
import FlatBuffers.Dl.AttributeF as AttributeF
import FlatBuffers.Dl.AttributeI as AttributeI
import FlatBuffers.Dl.AttributeType as AttributeType
import FlatBuffers.Dl.Dimension as Dimension
import FlatBuffers.Dl.DimensionValue as DimensionValue
import FlatBuffers.Dl.DimensionValueType as DimensionValueType
import FlatBuffers.Dl.Function as Function
import FlatBuffers.Dl.Graph as Graph
import FlatBuffers.Dl.Model as Model
import FlatBuffers.Dl.Node as Node
import FlatBuffers.Dl.OperatorSetId as OperatorSetId
import FlatBuffers.Dl.Tensor as Tensor
import FlatBuffers.Dl.TensorDataType as TensorDataType
import FlatBuffers.Dl.TensorShape as TensorShape
import FlatBuffers.Dl.TensorTypeAndShape as TensorTypeAndShape
import FlatBuffers.Dl.TypeInfo as TypeInfo
import FlatBuffers.Dl.TypeInfoValue as TypeInfoValue
import FlatBuffers.Dl.ValueInfo as ValueInfo
import mapping as mapping


def make_node(
    op_type: str,
    inputs: Sequence[str],
    outputs: Sequence[str],
    name: Optional[str] = None,
    doc_string: Optional[str] = None,
    domain: Optional[str] = None,
    **kwargs: Any,
) -> Node.NodeT:
    """Construct a Node.

    Args:
        op_type (string): The name of the operator to construct
        inputs (list of string): list of input names
        outputs (list of string): list of output names
        name (string, default None): optional unique identifier for NodeFbs
        doc_string (string, default None): optional documentation string for NodeFbs
        domain (string, default None): optional domain for NodeFbs.
            If it's None, we will just use default domain (which is empty)
        **kwargs (dict): the attributes of the node.  The acceptable values
            are documented in :func:`make_attribute`.

    Returns:
        Node
    """
    node = Node.NodeT()
    node.opType = op_type
    node.input = [*inputs]
    node.output = [*outputs]
    if name:
        node.name = name
    if doc_string:
        node.docString = doc_string
    if domain is not None:
        node.domain = domain
    if kwargs:
        node.attribute = [
            make_attribute(key, value)
            for key, value in sorted(kwargs.items())
            if value is not None
        ]
    return node


def make_operatorsetid(
    domain: str,
    version: int,
) -> OperatorSetId.OperatorSetIdT:
    """Construct an OperatorSetId.

    Args:
        domain (string): The domain of the operator set id
        version (integer): Version of operator set id
    Returns:
        OperatorSetId.OperatorSetIdT
    """
    operatorsetid = OperatorSetId.OperatorSetIdT()
    operatorsetid.domain = domain
    operatorsetid.version = version
    return operatorsetid


def make_graph_test_value(
    valuesForTest: Dict[str, Dict[str, np.ndarray]] = None, 
    exponents: Dict[str, List[int]] = None
) -> Tuple[Sequence[Tensor.TensorT], Sequence[Tensor.TensorT]]:
    """Construct the test_inputs_value, test_outputs_value of Graph

    Args:
        valuesForTest (Dict[str, Dict[str, np.ndarray]]): the test values used to compare accuracy.
                                                          The input format is as follows:
            {
                'inputs': {
                    'input_0_name': np.ndarray
                    ......
                    'input_n_name': np.ndarray
                },
                'outputs': {
                    'output_0_name': np.ndarray
                    ......
                    'output_n_name': np.ndarray
                },
            }

    Returns:
        Tuple[Sequence[Tensor.TensorT], Sequence[Tensor.TensorT]]
    """
    input_test_values = []
    output_test_values = []
    if valuesForTest is not None:
        # Get the input test value
        for input_name, input_value in valuesForTest.get('inputs', {}).items():
            var_exponents = exponents.get(input_name, [0])
            tensor_type = np_dtype_to_tensor_dtype(input_value.dtype)
            input_shape = input_value.shape
            if len(input_shape) > 1:
                input_value = input_value.flatten()
            if len(input_shape) == 0:
                input_shape = [1]
            input_value_fbs = make_tensor(input_name, tensor_type, input_shape, input_value.tobytes(), raw = True, exponents = var_exponents)
            input_test_values.append(input_value_fbs)

        # Get the output test value
        for output_name, output_value in valuesForTest.get('outputs', {}).items():
            var_exponents = exponents.get(output_name, [0])
            tensor_type = np_dtype_to_tensor_dtype(output_value.dtype)
            output_shape = output_value.shape
            if len(output_shape) > 1:
                output_value = output_value.flatten()
            if len(output_shape) == 0:
                output_shape = [1]
            output_value_fbs = make_tensor(output_name, tensor_type, output_shape, output_value.tobytes(), raw = True, exponents = var_exponents)
            output_test_values.append(output_value_fbs)

    return input_test_values, output_test_values


def make_graph(
    nodes: Sequence[Node.NodeT],
    name: str,
    inputs: Sequence[ValueInfo.ValueInfoT],
    outputs: Sequence[ValueInfo.ValueInfoT],
    initializer: Optional[Sequence[Tensor.TensorT]] = None,
    doc_string: Optional[str] = None,
    value_info: Optional[Sequence[ValueInfo.ValueInfoT]] = None,
    test_inputs_value: Optional[Sequence[Tensor.TensorT]] = None,
    test_outputs_value: Optional[Sequence[Tensor.TensorT]] = None,
) -> Graph.GraphT:
    """Construct a Graph

    Args:
        nodes: list of Node.NodeT
        name (string): graph name
        inputs: list of ValueInfo.ValueInfoT
        outputs: list of ValueInfo.ValueInfoT
        initializer: list of Tensor.TensorT
        doc_string (string): graph documentation
        value_info: list of ValueInfo.ValueInfoT
    Returns:
        Graph.GraphT
    """
    if initializer is None:
        initializer = []
    if value_info is None:
        value_info = []
    if test_inputs_value is None:
        test_inputs_value = []
    if test_outputs_value is None:
        test_outputs_value = []
    graph = Graph.GraphT()
    graph.node = [*nodes]
    graph.name = name
    graph.input = [*inputs]
    graph.output = [*outputs]
    graph.initializer = [*initializer]
    graph.valueInfo = [*value_info]
    if doc_string:
        graph.docString = doc_string
    graph.testInputsValue = [*test_inputs_value]
    graph.testOutputsValue = [*test_outputs_value]
    return graph


def make_model(graph: Graph.GraphT, **kwargs: Any) -> Model.ModelT:
    """Construct a Model

    Args:
        graph (Graph.GraphT): *make_graph* returns
        **kwargs: any attribute to add to the returned instance
    Returns:
        Model.ModelT
    """
    model = Model.ModelT()
    model.graph = graph

    opset_imports: Optional[Sequence[OperatorSetId.OperatorSetIdT]] = None
    opset_imports = kwargs.pop("opset_imports", None)  # type: ignore
    if opset_imports is not None:
        model.opsetImport = [*opset_imports]
    else:
        # Default import
        model.opsetImport = [OperatorSetId.OperatorSetIdT()]

    functions: Optional[Sequence[Function.FunctionT]] = None
    functions = kwargs.pop("functions", None)  # type: ignore
    if functions is not None:
        model.functions = [*functions]

    for k, v in kwargs.items():
        # TODO: Does this work with repeated fields?
        setattr(model, k, v)
    return model


def make_tensor(
    name: str, data_type: int, dims: Sequence[int], vals: Any, raw: bool = False, exponents: Sequence[int] = [0]
) -> Tensor.TensorT:
    """Make a Tensor with specified arguments.  If raw is False, this
    function will choose the corresponding proto field to store the
    values based on data_type. If raw is True, use "raw_data" proto
    field to store the values, and values should be of type bytes in
    this case.

    Args:
        name (string): tensor name
        data_type (int): a value such as TensorDataType.FLOAT
        dims (List[int]): shape
        vals: values
        raw (bool): if True, vals contains the serialized content of the tensor,
            otherwise, vals should be a list of values of the type defined by *data_type*

    Returns:
        Tensor.TensorT
    """
    tensor = Tensor.TensorT()
    tensor.dataType = data_type
    tensor.name = name

    if data_type == TensorDataType.TensorDataType().STRING and raw:
        raise TypeError("Can not use raw_data to store string type.")

    np_dtype = tensor_dtype_to_np_dtype(data_type)

    # Check number of vals specified equals tensor size
    expected_size = 1
    if raw:
        expected_size = np_dtype.itemsize

    if type(vals) is np.ndarray and len(vals.shape) > 1:
        vals = vals.flatten()
    for d in dims:
        expected_size *= d

    if len(vals) != expected_size:
        raise ValueError(
            f"Number of values does not match tensor's size. Expected {expected_size}, but it is {len(vals)}. "
        )

    if raw:
        tensor.rawData = vals
    else:
        if data_type == TensorDataType.TensorDataType().FLOAT16:
            raise TypeError("Don't support FLOAT16.")
        elif data_type == TensorDataType.TensorDataType().BOOL:
            vals = np.array(vals).astype(int)
        elif data_type == TensorDataType.TensorDataType().STRING:
            vals = np.array(vals).astype(bytes)
        field = tensor_dtype_to_field(data_type)
        # getattr(tensor, field).extend(vals)
        setattr(tensor, field, vals)
    tensor.dims = [*dims]
    tensor.exponents = [*exponents]
    return tensor


def _to_bytes(value: Union[str, bytes]) -> bytes:
    """Coerce a string (or bytes) value into UTF-8 bytes."""
    return value if isinstance(value, bytes) else value.encode("utf-8")


def _to_str(value: Union[str, bytes]) -> str:
    """Coerce a string (or UTF-8 bytes) value into str."""
    return value if isinstance(value, str) else value.decode("utf-8", errors="ignore")


def make_attribute(
    key: str,
    value: Any,
    doc_string: Optional[str] = None,
    attr_type: Optional[int] = None,
) -> Attribute.AttributeT:
    """Makes an Attribute based on the value type."""
    attr = Attribute.AttributeT()
    attr.name = key
    if doc_string:
        attr.docString = doc_string

    # Singular cases
    if isinstance(value, numbers.Integral):
        attr.i = AttributeI.AttributeIT()
        attr.i.i = int(value)
        attr.attrType = AttributeType.AttributeType().INT
    elif isinstance(value, numbers.Real):
        attr.f = AttributeF.AttributeFT()
        attr.f.f = float(value)
        attr.attrType = AttributeType.AttributeType().FLOAT
    elif isinstance(value, (str, bytes)):
        # Encode strings into utf-8
        attr.s = _to_bytes(value)
        attr.attrType = AttributeType.AttributeType().STRING
    elif isinstance(value, Tensor.TensorT):
        attr.t = value
        attr.attrType = AttributeType.AttributeType().TENSOR
    elif isinstance(value, Graph.GraphT):
        attr.g = value
        attr.attrType = AttributeType.AttributeType().GRAPH
    elif isinstance(value, TypeInfo.TypeInfoT):
        attr.tp = value
        attr.attrType = AttributeType.AttributeType().TYPE_FBS
    # Iterable cases
    elif isinstance(value, collections.abc.Iterable):
        value = list(value)
        if len(value) == 0 and attr_type is None:
            raise ValueError(
                f"Could not infer attribute `{key}` type from empty iterator"
            )
        if attr_type is None:
            types = {type(v) for v in value}
            for exp_t, exp_enum in (
                (numbers.Integral, AttributeType.AttributeType().INTS),
                (numbers.Real, AttributeType.AttributeType().FLOATS),
                ((str, bytes), AttributeType.AttributeType().STRINGS),
                (Tensor.TensorT, AttributeType.AttributeType().TENSORS),
                (Graph.GraphT, AttributeType.AttributeType().GRAPHS),
                (TypeInfo.TypeInfoT, AttributeType.AttributeType().TYPE_FBSS),
            ):
                if all(issubclass(t, exp_t) for t in types):  # type: ignore[arg-type]
                    attr_type = exp_enum
                    break
            if attr_type is None:
                raise ValueError(
                    "Could not infer the attribute type from the elements of the passed Iterable value."
                )

        if attr_type == AttributeType.AttributeType().INTS:
            attr.ints = [*value]
            attr.attrType = AttributeType.AttributeType().INTS
        elif attr_type == AttributeType.AttributeType().FLOATS:
            attr.floats = [*value]
            attr.attrType = AttributeType.AttributeType().FLOATS
        elif attr_type == AttributeType.AttributeType().STRINGS:
            attr.strings = [_to_bytes(v) for v in value]
            attr.attrType = AttributeType.AttributeType().STRINGS
        elif attr_type == AttributeType.AttributeType().TENSORS:
            attr.tensors = [*value]
            attr.attrType = AttributeType.AttributeType().TENSORS
        elif attr_type == AttributeType.AttributeType().GRAPHS:
            attr.graphs = [*value]
            attr.attrType = AttributeType.AttributeType().GRAPHS
        elif attr_type == AttributeType.AttributeType().TYPE_FBSS:
            attr.typeProtos = [*value]
            attr.attrType = AttributeType.AttributeType().TYPE_FBSS
        else:
            raise AssertionError()  # Should not reach since `ValueError` must be raised in attr_type checking
    else:
        raise TypeError(f"'{value}' is not an accepted attribute value.")

    if attr_type is not None and attr.attrType != attr_type:
        raise TypeError(
            f"Inferred attribute type ({attr.attrType}) mismatched with specified type ({attr_type})"
        )
    return attr


def get_attribute_value(attr: Attribute.AttributeT) -> Any:  # noqa: PLR0911
    if attr.refAttrName:
        raise ValueError(f"Cannot get value of reference attribute: {attr}")
    if attr.attrType == AttributeType.AttributeType.FLOAT:
        return attr.f.f
    if attr.attrType == AttributeType.AttributeType.INT:
        return attr.i.i
    if attr.attrType == AttributeType.AttributeType.STRING:
        return attr.s
    if attr.attrType == AttributeType.AttributeType.TENSOR:
        return attr.t
    if attr.attrType == AttributeType.AttributeType.GRAPH:
        return attr.g
    if attr.attrType == AttributeType.AttributeType.TYPE_FBS:
        return attr.tp
    if attr.attrType == AttributeType.AttributeType.FLOATS:
        return list(attr.floats)
    if attr.attrType == AttributeType.AttributeType.INTS:
        return list(attr.ints)
    if attr.attrType == AttributeType.AttributeType.STRINGS:
        return list(attr.strings)
    if attr.attrType == AttributeType.AttributeType.TENSORS:
        return list(attr.tensors)
    if attr.attrType == AttributeType.AttributeType.GRAPHS:
        return list(attr.graphs)
    if attr.attrType == AttributeType.AttributeType.TYPE_FBSS:
        return list(attr.typeProtos)
    if attr.attrType == AttributeType.AttributeType.UNDEFINED:
        return None
    raise ValueError(f"Unsupported ONNX attribute: {attr}")


def make_tensor_type_info(
    elem_type: int,
    shape: Optional[Sequence[Union[str, int, None]]],
    shape_denotation: Optional[List[str]] = None,
) -> TypeInfo.TypeInfoT:
    tensorShape = TensorShape.TensorShapeT()
    tensorShape.dim = []

    if shape is not None:
        if shape_denotation and len(shape_denotation) != len(shape):
            raise ValueError(
                "Invalid shape_denotation. Must be of the same length as shape."
            )

        for i, d in enumerate(shape):
            dimension_value = DimensionValue.DimensionValueT()
            tensorShape_dim = Dimension.DimensionT()
            if d is None:
                pass
            elif isinstance(d, int):
                dimension_value.dimType = DimensionValueType.DimensionValueType().VALUE
                dimension_value.dimValue = d
            elif isinstance(d, str):
                dimension_value.dimType = DimensionValueType.DimensionValueType().PARAM
                dimension_value.dimParam = d
            else:
                raise ValueError(
                    f"Invalid item in shape: {d}. Needs to be of int or str."
                )

            tensorShape_dim.value = dimension_value
            if shape_denotation:
                tensorShape_dim.denotation = shape_denotation[i]

            tensorShape.dim.append(tensorShape_dim)

    value = TensorTypeAndShape.TensorTypeAndShapeT()
    value.elemType = elem_type
    value.shape = tensorShape

    """Makes a Tensor TypeInfo.TypeInfoT based on the data type and shape."""
    type_info = TypeInfo.TypeInfoT()
    type_info.valueType = TypeInfoValue.TypeInfoValue().tensor_type
    type_info.value = value

    return type_info


def make_tensor_value_info(
    name: str,
    elem_type: int,
    shape: Optional[Sequence[Union[str, int, None]]],
    doc_string: str = "",
    shape_denotation: Optional[List[str]] = None,
    exponents: Sequence[int] = [0]
) -> ValueInfo.ValueInfoT:
    """Makes a ValueInfo.ValueInfoT based on the data type and shape."""
    value_info = ValueInfo.ValueInfoT()
    value_info.name = name
    if doc_string:
        value_info.docString = doc_string

    type_info = make_tensor_type_info(elem_type, shape, shape_denotation)
    value_info.valueInfoType = type_info
    value_info.exponents = [*exponents]
    return value_info

def aes_encryptor(data: bytes):
    key = bytearray(random.getrandbits(8) for _ in range(16))
    iv = bytearray(range(16, 32, 1))
    cipher = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
    encryptor = cipher.encryptor()
    ct_data = encryptor.update(data) + encryptor.finalize()
    
    hex_key = ''
    for _, byte in enumerate(key):
        hex_key += "0x{:02x}, ".format(byte)

    return ct_data, hex_key


def aes_decryptor(data: bytes, key: bytearray):
    iv = bytearray(range(16, 32, 1))
    cipher = Cipher(algorithms.AES(key), modes.CTR(iv), backend=default_backend())
    decryptor = cipher.decryptor()
    decrypted_data = decryptor.update(data) + decryptor.finalize()
    return decrypted_data

def save(
    model: Model.ModelT, file_path: str, encrypt_data: bool = False
):
    builder = flatbuffers.Builder(1024)
    builder.Finish(model.Pack(builder))
    data = builder.Output()
    key = None
    if encrypt_data:
        data, key = aes_encryptor(data)
    
    # file header 
    header = bytearray("EDL1", 'utf-8')
    header += struct.pack("I", 1 if encrypt_data else 0)
    header += struct.pack("I", len(data))
    with open(file_path, mode="wb") as f:
        f.write(header + data)
    return key


def load(file_path: str) -> Model.ModelT:
    buf = None
    with open(file_path, mode="rb") as f:
        buf = f.read()
    return Model.ModelT.InitFromPackedBuf(buf, 0)


def _sanitize_str(s: Union[str, bytes]) -> str:
    if isinstance(s, str):
        sanitized = s
    elif isinstance(s, bytes):
        sanitized = s.decode("utf-8", errors="ignore")
    else:
        sanitized = str(s)
    if len(sanitized) < 64:  # noqa: PLR2004
        return sanitized
    return sanitized[:64] + f"...<+len={(len(sanitized) - 64)}>"


def printable_attribute(
    attr: Attribute.AttributeT, subgraphs: bool = False
) -> Union[str, Tuple[str, List[Graph.GraphT]]]:
    content = []
    content.append(_to_str(attr.name))
    content.append("=")

    def str_float(f: float) -> str:
        # NB: Different Python versions print different numbers of trailing
        # decimals, specifying this explicitly keeps it consistent for all
        # versions
        return f"{f:.15g}"

    def str_int(i: int) -> str:
        return str(i)

    _T = TypeVar("_T")

    def str_list(str_elem: Callable[[_T], str], xs: Sequence[_T]) -> str:
        return "[" + ", ".join(map(str_elem, xs)) + "]"

    # for now, this logic should continue to work as long as we are running on a flatbuffers
    # implementation. If/when we switch to flatbuffers, we will need to use attr.type

    # To support printing subgraphs, if we find a graph attribute, print out
    # its name here and pass the graph itself up to the caller for later
    # printing.
    graphs = []
    if attr.i:
        content.append(str_int(attr.i.i))
    elif attr.f:
        content.append(str_float(attr.f.f))
    elif attr.s:
        # TODO: Bit nervous about Python 2 / Python 3 determinism implications
        content.append(repr(_sanitize_str(attr.s)))
    elif attr.t:
        if len(attr.t.dims) > 0:
            content.append("<Tensor>")
        else:
            # special case to print scalars
            field = tensor_dtype_to_field(attr.t.dataType)
            content.append(f"<Scalar Tensor {getattr(attr.t, field)}>")
    elif attr.g:
        content.append(f"<graph {_to_str(attr.g.name)}>")
        graphs.append(attr.g)
    elif attr.tp:
        content.append(f"<Type Info {attr.tp}>")
    elif attr.floats:
        content.append(str_list(str_float, attr.floats))
    elif attr.ints is not None:
        content.append(str_list(str_int, attr.ints))
    elif attr.strings:
        # TODO: Bit nervous about Python 2 / Python 3 determinism implications
        content.append(str(list(map(_sanitize_str, attr.strings))))
    elif attr.tensors:
        content.append("[<Tensor>, ...]")
    elif attr.typeProtos:
        content.append("[")
        for i, tp in enumerate(attr.typeProtos):
            comma = "," if i != len(attr.typeProtos) - 1 else ""
            content.append(f"<Type Info {tp}>{comma}")
        content.append("]")
    elif attr.graphs:
        content.append("[")
        for i, g in enumerate(attr.graphs):
            comma = "," if i != len(attr.graphs) - 1 else ""
            content.append(f"<graph {_to_str(g.name)}>{comma}")
        content.append("]")
        graphs.extend(attr.graphs)
    else:
        content.append("<Unknown>")
    if subgraphs:
        return " ".join(content), graphs
    return " ".join(content)


def printable_dim(dim: Dimension.DimensionT) -> str:
    if dim.value.dimType == DimensionValueType.DimensionValueType.UNKNOWN:
        return "?"
    elif dim.value.dimType == DimensionValueType.DimensionValueType.VALUE:
        which = dim.value.dimValue
    elif dim.value.dimType == DimensionValueType.DimensionValueType.PARAM:
        which = dim.value.dimParam
    return str(which)


def get_class_attribute_name(obj: Any, value: int) -> str:
    ret: str = ""
    # 获取类和实例的所有属性名字
    all_attributes = dir(obj)
    filtered_attributes = [attr for attr in all_attributes if not attr.startswith('__') and not attr.endswith('__')]
    for attr in filtered_attributes:
        if getattr(obj, attr) == value:
            ret = attr
            break

    return ret


def printable_type(t: TypeInfo.TypeInfoT) -> str:
    if t.valueType == TypeInfoValue.TypeInfoValue.tensor_type:
        s = get_class_attribute_name(TensorDataType.TensorDataType(), t.value.elemType)
        if t.value.shape is not None:
            if len(t.value.shape.dim):
                s += str(", " + "x".join(map(printable_dim, t.value.shape.dim)))
            else:
                s += ", scalar"
        return s  # type: ignore[no-any-return]
    if t.valueType == TypeInfoValue.TypeInfoValue.NONE:
        return ""
    return f"Unknown type {t.valueType}"


def printable_value_info(v: ValueInfo.ValueInfoT) -> str:
    s = f"%{_to_str(v.name)}"
    if v.valueInfoType:
        s = f"{s}[{printable_type(v.valueInfoType)}]"
    if v.exponents:
        s += f", exponents: {repr(v.exponents)}"
    return s


def printable_tensor(t: Tensor.TensorT) -> str:
    s = f"%{_to_str(t.name)}["
    s += get_class_attribute_name(TensorDataType.TensorDataType(), t.dataType)
    if t.dims is not None:
        if len(t.dims):
            s += str(", " + "x".join(map(str, t.dims)))
        else:
            s += ", scalar"
    s += "]"
    return s


def printable_tensor_value(t: Tensor.TensorT) -> str:
    s = f"%{_to_str(t.name)}, "
    if t.dims is not None:
        s += f"shape: {repr(t.dims)}, "
    if t.exponents is not None:
        s += f"exponents: {repr(t.exponents)}, "
    if t.docString is not None:
        s += f"docString: {t.docString}, "

    if t.floatData is not None:
        s += f"value: {repr(t.floatData)}"
    elif t.int32Data is not None:
        s += f"value: {repr(t.int32Data)}"
    elif t.stringData is not None:
        s += f"value: {repr(t.stringData)}"
    elif t.int64Data is not None:
        s += f"value: {repr(t.int64Data)}"
    elif t.rawData is not None:
        np.set_printoptions(threshold=np.inf)
        s += f"\nvalue: {repr(np.frombuffer(t.rawData, dtype=tensor_dtype_to_np_dtype(t.dataType)))}"
    elif t.doubleData is not None:
        s += f"value: {repr(t.doubleData)}"
    elif t.uint64Data is not None:
        s += f"value: {repr(t.uint64Data)}"

    return s


def printable_node(
    node: Node.NodeT, prefix: str = "", subgraphs: bool = False
) -> Union[str, Tuple[str, List[Graph.GraphT]]]:
    content = []
    if len(node.output):
        content.append(", ".join([f"%{_to_str(name)}" for name in node.output]))
        content.append("=")
    # To deal with nested graphs
    graphs: List[Graph.GraphT] = []
    printed_attrs = []
    if node.attribute:
        for attr in node.attribute:
            if subgraphs:
                printed_attr_subgraphs = printable_attribute(attr, subgraphs)
                if not isinstance(printed_attr_subgraphs[1], list):
                    raise TypeError(
                        f"printed_attr_subgraphs[1] must be an instance of {list}."
                    )
                graphs.extend(printed_attr_subgraphs[1])
                printed_attrs.append(printed_attr_subgraphs[0])
            else:
                printed = printable_attribute(attr)
                if not isinstance(printed, str):
                    raise TypeError(f"printed must be an instance of {str}.")
                printed_attrs.append(printed)
    printed_attributes = ", ".join(sorted(printed_attrs))
    printed_inputs = ", ".join([f"%{_to_str(name)}" for name in node.input])
    if node.attribute:
        content.append(f"{_to_str(node.opType)}[{printed_attributes}]({printed_inputs})")
    else:
        content.append(f"{_to_str(node.opType)}({printed_inputs})")
    if subgraphs:
        return prefix + " ".join(content), graphs
    return prefix + " ".join(content)


def printable_graph(graph: Graph.GraphT, prefix: str = "", print_initializer_value: bool = False, print_value_info: bool = False, print_test_value: bool = False) -> str:
    """Display a Graph as a string.

    Args:
        graph (Graph.GraphT): the graph to display
        prefix (string): prefix of every line

    Returns:
        string
    """
    content = []
    indent = prefix + "  "
    # header
    header = ["graph", _to_str(graph.name)]
    initializers = {t.name for t in graph.initializer}
    if len(graph.input):
        header.append("(")
        in_strs = []  # required inputs
        in_with_init_strs = (
            []
        )  # optional inputs with initializer providing default value
        for inp in graph.input:
            if inp.name not in initializers:
                in_strs.append(printable_value_info(inp))
            else:
                in_with_init_strs.append(printable_value_info(inp))
        if in_strs:
            content.append(prefix + " ".join(header))
            header = []
            for line in in_strs:
                content.append(prefix + "  " + line)
        header.append(")")

        if in_with_init_strs:
            header.append("optional inputs with matching initializers (")
            content.append(prefix + " ".join(header))
            header = []
            for line in in_with_init_strs:
                content.append(prefix + "  " + line)
            header.append(")")

        # from IR 4 onwards an initializer is not required to have a matching graph input
        # so output the name, type and shape of those as well
        if len(in_with_init_strs) < len(initializers):
            graph_inputs = {i.name for i in graph.input}
            init_strs = [
                printable_tensor(i)
                for i in graph.initializer
                if i.name not in graph_inputs
            ]
            header.append("initializers (")
            content.append(prefix + " ".join(header))
            header = []
            for line in init_strs:
                content.append(prefix + "  " + line)
            header.append(")")

    header.append("{")
    content.append(prefix + " ".join(header))
    graphs: List[Graph.GraphT] = []
    # body
    for node in graph.node:
        contents_subgraphs = printable_node(node, indent, subgraphs=True)
        if not isinstance(contents_subgraphs[1], list):
            raise TypeError(f"contents_subgraphs[1] must be an instance of {list}.")
        content.append(contents_subgraphs[0])
        graphs.extend(contents_subgraphs[1])
    # tail
    tail = ["return"]
    if len(graph.output):
        tail.append(", ".join([f"%{_to_str(out.name)}" for out in graph.output]))
    content.append(indent + " ".join(tail))
    # closing bracket
    content.append(prefix + "}")
    for g in graphs:
        content.append("\n" + printable_graph(g))

    if (print_initializer_value):
        initializers_value = [
                printable_tensor_value(i)
                for i in graph.initializer
            ]
        content.append("\n\n" + "initializers value:\n" + "\n".join(initializers_value))

    if (print_value_info):
        values_info = [
                printable_value_info(i)
                for i in graph.valueInfo
            ]
        content.append("\n\n" + "values info:\n" + "\n".join(values_info))

    if (print_test_value):
        test_inputs_value = [
                printable_tensor_value(i)
                for i in graph.testInputsValue
            ]
        content.append("\n\n" + "test inputs value:\n" + "\n".join(test_inputs_value))
        test_outputs_value = [
                printable_tensor_value(i)
                for i in graph.testOutputsValue
            ]
        content.append("\n\n" + "test outputs value:\n" + "\n".join(test_outputs_value))

    return "\n".join(content)


# Following functions are used for mapping
def tensor_dtype_to_np_dtype(tensor_dtype: int) -> np.dtype:
    """Convert a TensorDataType to corresponding numpy dtype. It can be used while making tensor.

    Args:
        tensor_dtype: TensorDataType

    Returns:
        numpy's data_type
    """
    return mapping.TENSOR_TYPE_MAP[tensor_dtype].np_dtype


def tensor_dtype_to_field(tensor_dtype: int) -> str:
    """Convert a TensorDataType to corresponding field name for storage. It can be used while making tensors.

    Args:
        tensor_dtype: TensorDataType

    Returns:
        field name
    """
    return mapping._STORAGE_TENSOR_TYPE_TO_FIELD[
        mapping.TENSOR_TYPE_MAP[tensor_dtype].storage_dtype
    ]


def np_dtype_to_tensor_dtype(np_dtype: np.dtype) -> int:
    """Convert a numpy's dtype to corresponding TensorDataType. It can be used while converting numpy arrays to tensors.

    Args:
        np_dtype: numpy's data_type

    Returns:
        TensorDataType
    """
    tensor_type = mapping._NP_TYPE_TO_TENSOR_TYPE.get(np_dtype)
    if tensor_type == None:
        return cast(int, TensorDataType.TensorDataType().UNDEFINED)
    else:
        return cast(int, tensor_type)
    # return cast(
    #     int,
    #     mapping._NP_TYPE_TO_TENSOR_TYPE[np_dtype],
    # )



from setuptools import find_packages, setup

ESP_PPQ_VERSION = "v0.2.1"

def readme():
    with open('README.md', encoding='utf-8') as f:
        content = f.read()
    return content

setup(author='esp-dl',
      author_email='dcp-ppq@sensetime.com',
      description='PPQ is an offline quantization tools',
      long_description=readme(),
      long_description_content_type='text/markdown',
      install_requires=open('requirements.txt').readlines(),
      python_requires='>=3.6',
      name='esp-ppq',
      package_dir={'ppq': 'ppq'},
      packages=find_packages(where='ppq'),
      classifiers=[
            'Development Status :: 3 - Alpha',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
      license='Apache License 2.0',
      include_package_data=True,
      version=ESP_PPQ_VERSION,
      zip_safe=False
    )



from ppq import *
from ppq.IR.morph import GraphFormatter
from ppq.api.interface import export_ppq_graph

# TEST CASE 1
graph = BaseGraph(name='Graph', built_from=NetworkFramework.ONNX)
graph.append_operation(operation=Operation(name='op1', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op2', op_type='Conv', attributes={}))
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op1'],
                          B=graph.operations['op2'])

processor = SearchableGraph(graph)
paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: False,
    ep_expr=lambda x: x.type == 'Conv',
    direction='down')

assert len(paths) == 1
path = paths[0]
assert path[0].name == 'op1'
assert path[1].name == 'op2'
assert path[-1].name == 'op2'
export_ppq_graph(graph=graph, platform=TargetPlatform.ONNX, graph_save_to='graph')

# TEST CASE 2
graph = BaseGraph(name='Graph', built_from=NetworkFramework.ONNX)
graph.append_operation(operation=Operation(name='op1', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op2', op_type='Add', attributes={}))
graph.append_operation(operation=Operation(name='op3', op_type='Relu', attributes={}))
graph.append_operation(operation=Operation(name='op4', op_type='Sigmoid', attributes={}))
graph.append_operation(operation=Operation(name='op5', op_type='Add', attributes={}))
graph.append_operation(operation=Operation(name='op6', op_type='Relu', attributes={}))
graph.append_operation(operation=Operation(name='op7', op_type='Sigmoid', attributes={}))
graph.append_operation(operation=Operation(name='op8', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op9', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op10', op_type='Conv', attributes={}))
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op1'],
                          B=graph.operations['op2'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op8'],
                          B=graph.operations['op2'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op2'],
                          B=graph.operations['op3'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op3'],
                          B=graph.operations['op4'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op9'],
                          B=graph.operations['op5'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op10'],
                          B=graph.operations['op5'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op5'],
                          B=graph.operations['op6'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op6'],
                          B=graph.operations['op7'])
processor = SearchableGraph(graph)
paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Sigmoid',
    direction='down')

assert len(paths) == 4
path = paths[0]
assert path[0].type == 'Conv'
assert path[1].type == 'Add'
assert path[-1].type == 'Sigmoid'

paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Relu',
    direction='down')

assert len(paths) == 4
path = paths[0]
assert path[0].type == 'Conv'
assert path[1].type == 'Add'
assert path[-1].type == 'Relu'

paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: False,
    ep_expr=lambda x: x.type == 'Relu',
    direction='down')
assert len(paths) == 0

graph.remove_operation(graph.operations['op1'])
paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Sigmoid',
    direction='down')
assert len(paths) == 3
path = paths[0]
assert path[0].type == 'Conv'
assert path[1].type == 'Add'
assert path[-1].type == 'Sigmoid'

opset = processor.opset_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Sigmoid',
    direction='down')
assert len(opset) == 9

graph.remove_operation(graph.operations['op2'])
graph.remove_operation(graph.operations['op3'])
graph.remove_operation(graph.operations['op4'])
opset = processor.opset_matching(
    sp_expr=lambda x: x.type == 'Conv',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Sigmoid',
    direction='down')
assert len(opset) == 5

processor = GraphFormatter(graph)
processor.delete_isolated()
assert len(graph.operations) == 0
export_ppq_graph(graph=graph, platform=TargetPlatform.ONNX, graph_save_to='graph')

# TEST CASE 3
graph = BaseGraph(name='Graph', built_from=NetworkFramework.ONNX)
graph.append_operation(operation=Operation(name='op1', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op2', op_type='Add', attributes={}))
graph.append_operation(operation=Operation(name='op3', op_type='Relu', attributes={}))
graph.append_operation(operation=Operation(name='op4', op_type='Sigmoid', attributes={}))
graph.append_operation(operation=Operation(name='op5', op_type='Add', attributes={}))
graph.append_operation(operation=Operation(name='op6', op_type='Relu', attributes={}))
graph.append_operation(operation=Operation(name='op7', op_type='Sigmoid', attributes={}))
graph.append_operation(operation=Operation(name='op8', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op9', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op10', op_type='Conv', attributes={}))
graph.append_operation(operation=Operation(name='op11', op_type='Conv', attributes={}))
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op1'],
                          B=graph.operations['op2'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op8'],
                          B=graph.operations['op2'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op2'],
                          B=graph.operations['op3'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op3'],
                          B=graph.operations['op4'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op9'],
                          B=graph.operations['op5'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op10'],
                          B=graph.operations['op5'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op5'],
                          B=graph.operations['op6'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op6'],
                          B=graph.operations['op7'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op6'],
                          B=graph.operations['op7'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op11'],
                          B=graph.operations['op1'])
graph.create_link_with_op(variable=graph.create_variable(),
                          A=graph.operations['op11'],
                          B=graph.operations['op10'])
processor = GraphFormatter(graph)
graph.mark_variable_as_graph_output(var=graph.operations['op6'].outputs[0])
processor.truncate_on_var(graph.operations['op1'].outputs[0], mark_as_output=True)
processor.delete_isolated()
assert len(graph.outputs) == 2
assert len(graph.operations) == 6

for var in graph.variables.copy():
    graph.insert_op_on_var(graph.create_operation(op_type='Test_1', attributes={}), var=var)
for var in graph.variables.copy():
    graph.insert_op_on_var(graph.create_operation(op_type='Test_2', attributes={}), var=var)

export_ppq_graph(graph=graph, platform=TargetPlatform.ONNX, graph_save_to='graph')
processor = SearchableGraph(graph)
paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Test_1',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Test_2',
    direction='down')
assert len(paths) == 8

paths = processor.path_matching(
    sp_expr=lambda x: x.type == 'Test_1',
    rp_expr=lambda x, y: True,
    ep_expr=lambda x: x.type == 'Test_2',
    direction='up')
assert len(paths) == 8



from .torch import (DEFAULT_BACKEND_TABLE, NXP_BACKEND_TABLE, EXTENSION_BACKEND_TABLE,
                    PPL_DSP_BACKEND_TABLE, PPL_GPU_BACKEND_TABLE, ONNX_BACKEND_TABLE,
                    ACADEMIC_BACKEND_TABLE, TorchBackendContext)



from ppq.lib import (
                     register_calibration_observer,
                     register_network_exporter,
                     register_network_parser,
                     register_network_quantizer,
                     register_operation_handler,
)

from .espdl_interface import (
                     espdl_quantize_onnx,
                     espdl_quantize_torch,
                     generate_test_value,
                     get_target_platform,
)
from .fsys import (
                     compare_cosine_similarity_between_results,
                     create_dir,
                     dump_internal_results,
                     dump_to_file,
                     load_calibration_dataset,
                     load_from_file,
                     split_result_to_directory,
)
from .interface import (
                     DISABLE_CUDA_KERNEL,
                     ENABLE_CUDA_KERNEL,
                     UnbelievableUserFriendlyQuantizationSetting,
                     dispatch_graph,
                     dump_torch_to_onnx,
                     empty_ppq_cache,
                     export,
                     export_ppq_graph,
                     format_graph,
                     load_caffe_graph,
                     load_graph,
                     load_native_graph,
                     load_onnx_graph,
                     load_torch_model,
                     manop,
                     quantize,
                     quantize_caffe_model,
                     quantize_native_model,
                     quantize_onnx_model,
                     quantize_torch_model,
)
from .setting import (
                     ActivationQuantizationSetting,
                     BiasCorrectionSetting,
                     BlockwiseReconstructionSetting,
                     ChannelSplitSetting,
                     DispatchingTable,
                     EqualizationSetting,
                     GraphFormatSetting,
                     LSQSetting,
                     ParameterQuantizationSetting,
                     QuantizationFusionSetting,
                     QuantizationSetting,
                     QuantizationSettingFactory,
                     SSDEqualizationSetting,
                     TemplateSetting,
                     WeightSplitSetting,
)



import torch

from ppq.IR import Variable
from ppq.lib import LinearQuantizationConfig
from ppq.quantization.observer import TorchIsotoneObserver, TorchMinMaxObserver
from ppq.quantization.qfunction import PPQLinearQuantFunction

from ppq import QuantizationStates

TQC = LinearQuantizationConfig()
v = Variable(name='TestVariable')

for i in range(10000):
    isotone_observer = TorchIsotoneObserver(v, TQC)
    minmax_observer  = TorchMinMaxObserver(v, TQC)

    TQC.state = QuantizationStates.INITIAL
    value = torch.softmax(torch.rand(size=[1, 10]), dim=-1)
    value, _ = torch.sort(value, dim=-1)
    isotone_observer.observe(value)
    isotone_observer.render_quantization_config()

    isotone_quant = PPQLinearQuantFunction(value, TQC)
    o = torch.argmax(value, dim=-1)
    q = torch.argmax(isotone_quant, dim=-1)
    isotone_error_num = torch.sum(o != q).item()
    isotone_scale     = TQC.scale

    TQC.state = QuantizationStates.INITIAL
    minmax_observer.observe(value)
    minmax_observer.render_quantization_config()

    minmax_quant = PPQLinearQuantFunction(value, TQC)
    o = torch.argmax(value, dim=-1)
    q = torch.argmax(minmax_quant, dim=-1)
    minmax_error_num = torch.sum(o != q).item()
    minmax_scale     = TQC.scale
    
    if not isotone_error_num <= minmax_error_num:
        print(isotone_observer.s_candidates)
        print(isotone_error_num, minmax_error_num)
        print(value)
        print(isotone_quant, isotone_scale)
        print(minmax_quant, minmax_scale)
        raise Exception('Test Failed.')


TQC = LinearQuantizationConfig(symmetrical=False)
v = Variable(name='TestVariable')

for i in range(10000):
    isotone_observer = TorchIsotoneObserver(v, TQC)
    minmax_observer  = TorchMinMaxObserver(v, TQC)

    TQC.state = QuantizationStates.INITIAL
    value = torch.softmax(torch.rand(size=[1, 10]), dim=-1)
    value, _ = torch.sort(value, dim=-1)
    isotone_observer.observe(value)
    isotone_observer.render_quantization_config()

    isotone_quant = PPQLinearQuantFunction(value, TQC)
    o = torch.argmax(value, dim=-1)
    q = torch.argmax(isotone_quant, dim=-1)
    isotone_error_num = torch.sum(o != q).item()
    isotone_scale     = TQC.scale

    TQC.state = QuantizationStates.INITIAL
    minmax_observer.observe(value)
    minmax_observer.render_quantization_config()

    minmax_quant = PPQLinearQuantFunction(value, TQC)
    o = torch.argmax(value, dim=-1)
    q = torch.argmax(minmax_quant, dim=-1)
    minmax_error_num = torch.sum(o != q).item()
    minmax_scale     = TQC.scale
    
    if not isotone_error_num <= minmax_error_num:
        print(isotone_observer.s_candidates)
        print(isotone_error_num, minmax_error_num)
        print(value)
        print(isotone_quant, isotone_scale)
        print(minmax_quant, minmax_scale)
        raise Exception('Test Failed.')


from typing import Iterable
from ppq.IR.base.graph import BaseGraph

from ppq.executor import BaseGraphExecutor
from ppq.IR import BaseGraph

from .base import QuantizationOptimizationPass


class ExtensionPass(QuantizationOptimizationPass):
    """ExtensionPass 并没有什么用，它就是告诉你你可以像这样写一个 pass。 你可以直接改写 ExtensionPass
    的逻辑来实现你的功能，并将修改后的代码提交到 github.

    不过比较我们已经为 ExtensionPass 创建了一个 TemplateSetting 用来给它传递参数
        你可以去 ppq.api.setting.py 里面找到它

    There is nothing in ExtensionPass, it is literally an empty pass,
        -- just show you how to create your own pass.

    A TemplateSetting class has been created for passing parameter to this pass.
        You can find it in ppq.api.setting.py

    You can overwrite logic inside this pass.
    """
    def __init__(self, parameter: str) -> None:
        self.parameter = parameter
        super().__init__(name='PPQ Extension Pass')

    def optimize(
        self,
        graph: BaseGraph,
        dataloader: Iterable,
        executor: BaseGraphExecutor,
        **kwargs
    ) -> None:
        assert isinstance(graph, BaseGraph)

        print('You are invoking Extension Pass now.')



from ppq import *
from ppq.IR.morph import GraphMerger
from ppq.api import *
import torch

graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Matmul', name='matmul', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Add', name='add', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphMerger(graph)
processor.fuse_gemm()

assert len(graph.operations) == 1
assert len(graph.operations['matmul'].inputs) == 3
assert graph.operations['matmul'].type == 'Gemm'

graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Matmul', name='matmul', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
test = \
graph.create_operation(op_type='Test', name='test', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[], outputs=[graph.create_variable()])
graph.create_operation(op_type='Add', name='add', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], test.outputs[0]], 
                       outputs=[graph.create_variable()])
processor = GraphMerger(graph)
processor.fuse_gemm()

assert len(graph.operations) == 3
assert len(graph.operations['matmul'].inputs) == 2
assert graph.operations['matmul'].type == 'Gemm'


graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Matmul', name='matmul', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Add', name='add', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[1, ]))], 
                       outputs=[graph.create_variable()])
processor = GraphMerger(graph)
processor.fuse_gemm()

assert len(graph.operations) == 2
assert len(graph.operations['matmul'].inputs) == 2
assert graph.operations['matmul'].type == 'Gemm'


from ppq import *
from ppq.IR.morph import GraphMerger
from ppq.api import *
import torch

graph = BaseGraph(name='test', built_from=NetworkFramework.ONNX)
matmul = \
graph.create_operation(op_type='Matmul', name='matmul', 
                       platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[graph.create_variable(), graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, 10]))],
                       outputs=[graph.create_variable()])
graph.create_operation(op_type='Relu', name='relu', platform=TargetPlatform.UNSPECIFIED, 
                       inputs=[matmul.outputs[0], graph.create_variable(is_parameter=True, value=torch.zeros(size=[10, ]))], 
                       outputs=[graph.create_variable()])
processor = QuantableGraph(graph)
processor.quantize_operation('matmul', target_platform=TargetPlatform.PPL_CUDA_INT8)
processor.quantize_operation('relu', target_platform=TargetPlatform.PPL_CUDA_INT8)



from ppq import *
from ppq.api import *
import torch

# 创建计算图
graph = BaseGraph(name='Created Graph', built_from=NetworkFramework.NATIVE)
op1 = graph.create_operation(
    op_type='Gemm', name='Gemm1',
    inputs=[graph.create_variable(), 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[128, 1024]).cuda() * 100), 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[1024]).cuda() * 100)],
    outputs=[graph.create_variable()])

op2 = graph.create_operation(
    op_type='Gemm', name='Gemm2',
    inputs=[op1.outputs[0], 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[1024, 128]).cuda()), 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[128]).cuda())],
    outputs=[graph.create_variable()])

op3 = graph.create_operation(
    op_type='Gemm', name='Gemm3', attributes={'transB': 1},
    inputs=[op2.outputs[0], 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[1024, 128]).cuda() * 500), 
            graph.create_variable(is_parameter=True, value=torch.rand(size=[1024]).cuda() * 0)],
    outputs=[graph.create_variable()])

graph.mark_variable_as_graph_input(op1.inputs[0])
graph.mark_variable_as_graph_output(op3.outputs[0])

inputs   = [torch.rand(size=[8, 128]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=1000, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7

import torch
class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        with torch.no_grad():
            self.conv1 = torch.nn.Conv2d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            self.conv2 = torch.nn.Conv2d(in_channels=32, out_channels=32, groups=16, kernel_size=3, stride=1, padding=1, bias=False)
            self.conv3 = torch.nn.Conv2d(in_channels=32, out_channels=8, groups=8, kernel_size=5, stride=1, padding=2)
            self.convtranspose1 = torch.nn.ConvTranspose2d(in_channels=8, out_channels=32, kernel_size=5, stride=1, padding=2)
            self.convtranspose2 = torch.nn.ConvTranspose2d(in_channels=32, out_channels=32, groups=32, kernel_size=3, stride=2, bias=False)
            self.convtranspose3 = torch.nn.ConvTranspose2d(in_channels=32, out_channels=8, groups=1, kernel_size=1)
            
            self.conv1.bias.copy_(torch.rand_like(self.conv1.bias))
            self.conv3.bias.copy_(torch.rand_like(self.conv3.bias))
            self.convtranspose1.bias.copy_(torch.rand_like(self.convtranspose1.bias))
            self.convtranspose3.bias.copy_(torch.rand_like(self.convtranspose3.bias))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.convtranspose1(x)
        x = self.convtranspose2(x)
        x = self.convtranspose3(x)
        return x

model = MyModel().cuda()
dump_torch_to_onnx(model=model, onnx_export_file='model.onnx', input_shape=[1, 8, 96, 96])
graph = load_onnx_graph(onnx_import_file='model.onnx')

inputs   = [torch.rand(size=[1, 8, 96, 96]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=10, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7


class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        with torch.no_grad():
            self.conv1 = torch.nn.Conv3d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            self.conv2 = torch.nn.Conv3d(in_channels=32, out_channels=32, groups=16, kernel_size=3, stride=1, padding=1, bias=False)
            self.conv3 = torch.nn.Conv3d(in_channels=32, out_channels=8, groups=8, kernel_size=5, stride=1, padding=2)
            self.conv4 = torch.nn.Conv3d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)

            self.conv1.bias.copy_(torch.rand_like(self.conv1.bias))
            self.conv3.bias.copy_(torch.rand_like(self.conv3.bias))
            self.conv4.bias.copy_(torch.rand_like(self.conv4.bias))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        return x

model = MyModel().cuda()
dump_torch_to_onnx(model=model, onnx_export_file='model.onnx', input_shape=[1, 8, 16, 96, 96])
graph = load_onnx_graph(onnx_import_file='model.onnx')

inputs   = [torch.rand(size=[1, 8, 16, 96, 96]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=10, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7


class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        with torch.no_grad():
            self.conv1 = torch.nn.Conv1d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            self.conv2 = torch.nn.Conv1d(in_channels=32, out_channels=32, groups=16, kernel_size=3, stride=1, padding=1, bias=False)
            self.conv3 = torch.nn.Conv1d(in_channels=32, out_channels=8, groups=8, kernel_size=5, stride=1, padding=2)
            self.conv4 = torch.nn.Conv1d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            
            self.conv1.bias.copy_(torch.rand_like(self.conv1.bias))
            self.conv3.bias.copy_(torch.rand_like(self.conv3.bias))
            self.conv4.bias.copy_(torch.rand_like(self.conv4.bias))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        return x

model = MyModel().cuda()
dump_torch_to_onnx(model=model, onnx_export_file='model.onnx', input_shape=[1, 8, 96])
graph = load_onnx_graph(onnx_import_file='model.onnx')

inputs   = [torch.rand(size=[1, 8, 96]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=10, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7


class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        with torch.no_grad():
            self.conv1 = torch.nn.ConvTranspose1d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            self.conv2 = torch.nn.ConvTranspose1d(in_channels=32, out_channels=32, groups=16, kernel_size=3, stride=1, padding=1, bias=False)
            self.conv3 = torch.nn.ConvTranspose1d(in_channels=32, out_channels=8, groups=8, kernel_size=5, stride=1, padding=2)
            self.conv4 = torch.nn.ConvTranspose1d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)

            self.conv1.bias.copy_(torch.rand_like(self.conv1.bias))
            self.conv3.bias.copy_(torch.rand_like(self.conv3.bias))
            self.conv4.bias.copy_(torch.rand_like(self.conv4.bias))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        return x

model = MyModel().cuda()
dump_torch_to_onnx(model=model, onnx_export_file='model.onnx', input_shape=[1, 8, 96])
graph = load_onnx_graph(onnx_import_file='model.onnx')

inputs   = [torch.rand(size=[1, 8, 96]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=10, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7

class MyModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        with torch.no_grad():
            self.conv1 = torch.nn.ConvTranspose3d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)
            self.conv2 = torch.nn.ConvTranspose3d(in_channels=32, out_channels=32, groups=16, kernel_size=3, stride=1, padding=1, bias=False)
            self.conv3 = torch.nn.ConvTranspose3d(in_channels=32, out_channels=8, groups=8, kernel_size=5, stride=1, padding=2)
            self.conv4 = torch.nn.ConvTranspose3d(in_channels=8, out_channels=32, kernel_size=3, stride=2, padding=1)

            self.conv1.bias.copy_(torch.rand_like(self.conv1.bias))
            self.conv3.bias.copy_(torch.rand_like(self.conv3.bias))
            self.conv4.bias.copy_(torch.rand_like(self.conv4.bias))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        return x

model = MyModel().cuda()
dump_torch_to_onnx(model=model, onnx_export_file='model.onnx', input_shape=[1, 8, 8, 8, 8])
graph = load_onnx_graph(onnx_import_file='model.onnx')

inputs   = [torch.rand(size=[1, 8, 8, 8, 8]) for _ in range(32)]
executor = TorchExecutor(graph=graph)
b_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
b_outputs = torch.cat(b_outputs)

from ppq.quantization.optim import LayerwiseEqualizationPass
LayerwiseEqualizationPass(iterations=10, including_bias=True, including_act=True).optimize(
    graph=graph, dataloader=inputs, executor=executor, collate_fn=lambda x: x.cuda())
p_outputs = [executor.forward(inputs=t.cuda())[0].unsqueeze(0) for t in inputs]
p_outputs = torch.cat(p_outputs)
from ppq.quantization.measure import torch_snr_error
assert torch_snr_error(b_outputs, p_outputs).item() < 1e-7


import torch

def torch_mean_square_error(y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str='mean') -> torch.Tensor:
    """
    Compute mean square error between y_pred(tensor) and y_real(tensor)
    
    MSE error can be calcualted as following equation:
    
        MSE(x, y) = (x - y) ^ 2
    
    if x and y are matrixs, MSE error over matrix should be the mean value of MSE error over all elements.
    
        MSE(X, Y) = mean((X - Y) ^ 2)

    By this equation, we can easily tell that MSE is an symmtrical measurement:
        MSE(X, Y) == MSE(Y, X)
        MSE(0, X) == X ^ 2

    Args:
        y_pred (torch.Tensor): _description_
        y_real (torch.Tensor): _description_
        reduction (str, optional): _description_. Defaults to 'mean'.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        torch.Tensor: _description_
    """
    if y_pred.shape != y_real.shape:
        raise ValueError(f'Can not compute mse loss for tensors with different shape. '
            f'({y_pred.shape} and {y_real.shape})')
    reduction = str(reduction).lower()

    if y_pred.ndim == 1:
        y_pred = y_pred.unsqueeze(0)
        y_real = y_real.unsqueeze(0)

    diff = torch.pow(y_pred - y_real, 2).flatten(start_dim=1)
    mse  = torch.mean(diff, dim=-1)

    if reduction == 'mean':
        return torch.mean(mse)
    elif reduction == 'sum':
        return torch.sum(mse)
    elif reduction == 'none':
        return mse
    else:
        raise ValueError(f'Unsupported reduction method.')

def torch_snr_error(y_pred: torch.Tensor, y_real: torch.Tensor, reduction: str='mean') -> torch.Tensor:
    """
    Compute SNR between y_pred(tensor) and y_real(tensor)
    
    SNR can be calcualted as following equation:
    
        SNR(pred, real) = (pred - real) ^ 2 / (real) ^ 2
    
    if x and y are matrixs, SNR error over matrix should be the mean value of SNR error over all elements.
    
        SNR(pred, real) = mean((pred - real) ^ 2 / (real) ^ 2)

    Args:
        y_pred (torch.Tensor): _description_
        y_real (torch.Tensor): _description_
        reduction (str, optional): _description_. Defaults to 'mean'.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        torch.Tensor: _description_
    """
    if y_pred.shape != y_real.shape:
        raise ValueError(f'Can not compute snr loss for tensors with different shape. '
            f'({y_pred.shape} and {y_real.shape})')
    reduction = str(reduction).lower()

    if y_pred.ndim == 1:
        y_pred = y_pred.unsqueeze(0)
        y_real = y_real.unsqueeze(0)

    y_pred = y_pred.flatten(start_dim=1)
    y_real = y_real.flatten(start_dim=1)

    noise_power  = torch.pow(y_pred - y_real, 2).sum(dim=-1)
    signal_power = torch.pow(y_real, 2).sum(dim=-1)
    snr = (noise_power) / (signal_power + 1e-7)

    if reduction == 'mean':
        return torch.mean(snr)
    elif reduction == 'sum':
        return torch.sum(snr)
    elif reduction == 'none':
        return snr
    else:
        raise ValueError(f'Unsupported reduction method.')
